import tornado
from tornado.httpclient import AsyncHTTPClient, HTTPError
from tornado import gen
from urllib.parse import urljoin
import concurrent.futures
from datetime import datetime, timedelta
import xml.etree.ElementTree as ET
import logging
import asyncio

executor = concurrent.futures.ThreadPoolExecutor(4)


def __get_rate_from_xml(xml, currency_code):
    xml = ET.fromstring(xml)

    currencies = xml.findall('Currency')

    f = filter(lambda x: x.attrib['CurrencyCode'] == currency_code, currencies)
    if f:
        f = f[0]
        return float(f.find('BanknoteSelling').text)



async def get_rate(currency_code, date=None):
    """
    yield currency_helper.get_rate('USD', datetime(2016,05,30))
    """

    base_url = 'http://www.tcmb.gov.tr/kurlar/'
    if not date:
        url = 'today.xml'
    else:
        url = '%s/%s.xml' % (date.strftime('%Y%m'), date.strftime('%d%m%Y'))

    request = tornado.httpclient.HTTPRequest(
        method='GET',
        url=urljoin(
            base_url,
            url
        )
    )

    logging.info(request.url)

    http_client = AsyncHTTPClient()
    try:
        response = await http_client.fetch(request)
        rate = await asyncio.futures.wrap_future(
            executor.submit(__get_rate_from_xml, response.body, currency_code))
        return rate
    except HTTPError as e:
        if e.response.code == 404:
            date = date + timedelta(days=-1)
            new_rate = await get_rate(currency_code, date)
            raise gen.Return(new_rate)

        raise e



import hashlib
from bson import ObjectId
import base64


class CodeGenerator(object):

    def __init__(self):
        pass

    def shortcode(self, text):
        """
        Our main shortening function. The rationale here is that
        we are relying on the fact that for similarly sized inputs
        such as URLs the potential for collision in the 32 last bits
        of the MD5 hash is rather unlikely.

        The following things happen, in order:

        * compute the md5 digest of the given source
        * extract the lower 4 bytes
        * base64 encode the result
        * remove trailing padding if it exists

        Of course, should a collision happen, we will evict the previous
        key.

        """
        return base64.b64encode(
            hashlib.md5(text.encode()).hexdigest()[-6:]
        ).replace('=', '').replace('/', '_')

    def coupon_code(self):
        _id = ObjectId()
        code = self.shortcode(str(_id))
        code = code.upper()
        code = code.replace('I', '1')
        code = code.replace('O', '2')
        code = code.replace('0', '3')
        code = code.replace('=', '4')
        code = code.replace('+', '5')
        code = code.replace('_', '6')

        return code



import user_agents
import logging
from urllib.parse import urlparse

def create_log(request, **kwargs):
    log = {}

    ip_address = request.headers.get(
        'X-Forwarded-For', request.remote_ip)
    ip_address = ip_address.split(',')[-1].strip()
    ip_address = request.headers.get('X-Real-Ip', ip_address)

    log['useragent'] = kwargs.get('useragent', request.headers.get('User-Agent'))
    log['referer'] = kwargs.get('referer', request.headers.get('Referer'))
    log['ip'] = kwargs.get('ip', ip_address)

    try:
        useragent = str(user_agents.parse(log.get('useragent')))
        log['useragent'] = useragent
    except Exception as ex:
        logging.error(ex)
        pass

    referer = None
    try:
        referer = urlparse(log.get('referer'))
        referer_dict = {
            'netloc': referer.netloc,
            'path': referer.path,
            'params': referer.params,
            'query': referer.query,
            'fragment': referer.fragment
        }
        log['referer'] = referer_dict
    except:
        pass

    return log





import logging

logger = logging.getLogger("event_hub")
hub = None

def init_event_hub(app):
    global hub
    hub = EventHub(app)

class Event(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

class EventHub(object):

    def __init__(self, app):
        self._app = app
        self._handlers = {}

    def subscribe(self, event_type):
        if event_type not in self._handlers:
            self._handlers[event_type] = set()

        def wrapper(f):
            self._handlers[event_type].add(f)
            return f

        return wrapper

    async def dispatch(self, event):
        logger.info("Event arrived <{}>".format(str(type(event))))

        handlers = self._handlers.get(type(event), [])
        for handler in handlers:
            await handler(event)





# coding=utf-8


def mask(card_number):
    """ Masks the credit card number """
    card_number = card_number[:6] + \
        card_number[-4:].rjust(len(card_number) - 6, "*")

    return card_number


def clear(card_info):
    """Clears the credit card number"""
    card_number = card_info.get('card_number', None)
    if card_number:
        card_info['card_number'] = mask(
            card_number
        )
        del card_info['cvv']




import pytest

from bson import ObjectId
from motor.motor_asyncio import AsyncIOMotorClient

from quark_utilities import mongol
from quark_utilities.mongol import Repository

test_collection_name = "test_col"


test_data = [
    {
        "_id": ObjectId(),
        "field_1": "value_1",
        "field_2": "value_2",
        "field_3": "value_3",
        "field_4": "value_4"
    },
    {
        "_id": ObjectId(),
        "field_1": "value_5",
        "field_2": "value_6",
        "field_3": "value_7",
        "field_4": "value_8"
    }
]


@pytest.mark.asyncio
async def test_maybe_object_id_should_return_object_id():
    object_id = ObjectId()
    string_id = str(ObjectId())
    not_object_id = "23124354657"

    assert object_id == mongol.maybe_object_id(object_id)
    assert string_id == str(mongol.maybe_object_id(string_id))
    assert not_object_id == mongol.maybe_object_id(not_object_id)


@pytest.mark.asyncio
async def resource_test():
    collection = AsyncIOMotorClient("localhost").test[test_collection_name]
    test_repo = Repository(collection)

    for data in test_data:
        await test_repo.create(data)

    #: Test Searchs
    assert await test_repo.find()["_id"] is not None, "Find one"
    assert len(await test_repo.query()) == 2, "Query without param"
    assert await test_repo.find(test_data[0]["_id"]) is not None, "Find by ID"

    #: Test update
    test_record_id = test_data[0]["_id"]
    founded = await test_repo.find(test_record_id)
    founded["updated_field"] = True
    await test_repo.update(test_record_id, founded)
    assert await test_repo.find(test_record_id)["updated_field"] is True

    #: Test Delete
    await test_repo.delete(test_record_id)
    founded = await test_repo.find(test_record_id, include_deleted=True)
    assert founded["is_deleted"] is True
    assert await test_repo.find(test_record_id) is None

    await test_repo.collection.drop()


async def load_test_data(collection):
    for data in test_data:
        await collection.create(data)


def get_db():
    return AsyncIOMotorClient("localhost").test



# coding=utf8
from asyncio import futures

import functools
import geoip2.database
import concurrent.futures
from tornado import gen
import os
current_dir = os.path.dirname(os.path.realpath(__file__))

# This creates a Reader object. You should use the same object
# across multiple requests as creation of it is expensive.
reader = geoip2.database.Reader(
    os.path.join(
        current_dir,
        '../',
        'app_data/GeoIP2-City.mmdb'
    )
)

executor = concurrent.futures.ThreadPoolExecutor(2)


async def get_location(ip_address):
    '''
    Ip adresinin ülkesini getirmek için
    '''
    response = await futures.wrap_future(executor.submit(
        functools.partial(reader.city, ip_address)
    ))

    if not response:
        return None

    return {
        'country': response.country.name,
        'city': response.city.name,
        'latitude': response.location.latitude,
        'longitude': response.location.longitude
    }



from setuptools import setup


setup(
    name='QuarkUtilities',
    version="0.0.11",
    license='BSD',
    author='Ali Arda Orhan',
    author_email='arda.orhan@dogantv.com.tr',
    description='Quark-CMS Utilities.',
    long_description=__doc__,
    packages=['quark_utilities'],
    platforms='any',
    install_requires=[
        'FlaskyTornado==0.0.29',
        'PyJWT==1.4.2',
        'jsonschema==2.6.0',
        'pymongo==3.4.0',
        'user_agents==1.1.0',
        "appdirs==1.4.3",
        "cffi==1.10.0",
        "cryptography==1.7.2"
        "FlaskyTornado==0.0.29",
        "geoip2==2.5.0",
        "idna==2.5",
        "Jinja2==2.9.6",
        "jsonschema==2.6.0",
        "jwt==0.5.2",
        "MarkupSafe==1.0",
        "maxminddb==1.3.0",
        "motor==1.1",
        "numpy==1.12.1",
        "olefile==0.44",
        "packaging==16.8",
        "Pillow==4.1.1",
        "pyasn1==0.2.3",
        "pycparser==2.17",
        "pymongo==3.4.0",
        "pyparsing==2.2.0",
        "python-dateutil==2.6.0",
        "requests==2.14.2",
        "scikit-learn==0.18.1",
        "six==1.10.0",
        "sklearn==0.0",
        "tornado==4.4.2",
        "ua-parser==0.7.3",
        "user-agents==1.1.0",
        "Wand==0.4.4"
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



import datetime
import logging
import json

from flasky.errors import FlaskyTornError, BadRequestError
from jsonschema import Draft4Validator, FormatChecker

logger = logging.getLogger("cms_utilities.json_schema")

__all__ = ["JSONSchemaValidator"]


class JSONSchemaValidationError(FlaskyTornError):

    def __init__(self, message=None, reason=None):
        super().__init__(status_code=400, message=message)
        self.reason = reason


class JSONSchemaValidator(object):

    def __init__(self, app):
        self._app = app
        self.init_app(app)

    def init_app(self, app):
        app.before_request(self.append_validator)
        app.before_request(self.validate_schema)

        @app.error_handler(JSONSchemaValidationError)
        async def handle_json_err(handler, err, endpoint_def):
            await self.handle_error(handler, err, endpoint_def)

    async def handle_error(self, handler, err, endpoint_def):
        handler.write(json.dumps({
                    "error": err.reason
                }))
        handler.set_status(400)

    def append_validator(self, handler, method_definition):
        if 'json_schema' not in method_definition:
            return
        schema = method_definition["json_schema"]
        setattr(handler, "validate_body", self.validator_func_factory(schema))

    def validator_func_factory(self, schema):
        def validator(handler):
            body = handler.body_as_json()
            return self.do_validation(body, schema)
        return validator

    def do_validation(self, body, schema):
        validator = Draft4Validator(
            schema,
            types={"datetime": datetime.datetime},
            format_checker=FormatChecker()
        )

        return sorted(validator.iter_errors(body), key=lambda e: e.path)

    def validate_schema(self, handler, method_definition):
        if 'json_schema' not in method_definition:
            return

        body = handler.body_as_json(throw_exc=True)

        schema = method_definition['json_schema']
        errors = self.do_validation(body, schema)
        if errors:
            raise JSONSchemaValidationError(
                message="Validation Error Occured", reason=format_validation_errors(errors))


def get_human_readable(error):
    path = ".".join(list(error.path))
    if not path:
        path = "root"
    return path, error.message


def format_validation_errors(errors):
    _errors = {}
    for error in errors:
        path, message = get_human_readable(error)
        _errors[path] = message
    return _errors



from bson import ObjectId
from flasky.errors import FlaskyTornError

__all__ = ["maybe_object_id", "Repository", "create_id"]


class RecordNotExistsError(FlaskyTornError):
    def __init__(self, record_type, record_id):
        super().__init__(status_code=400)


def maybe_object_id(maybe_id):
    if isinstance(maybe_id, ObjectId):
        return maybe_id
    elif not ObjectId.is_valid(maybe_id):
        return maybe_id
    else:
        return ObjectId(maybe_id)

def create_id():
    return ObjectId()

class Repository(object):

    def __init__(self, collection):
        self.collection = collection

    @property
    def collection_name(self):
        return self.collection.name

    async def query(self, select=None, where=None,
                    sort=None, skip=None, limit=None):
        cur = self.collection.find(where, select)

        if skip:
            cur.skip(skip)

        if sort:
            cur.sort(sort)

        if limit:
            cur.limit(limit)

        records = await cur.to_list(length=None)
        return [record
                for record in records
                if not record.get("is_deleted", False)]

    async def save(self, data):
        if not data.get("_id", None):
            # ~append id if it not exists
            data["_id"] = ObjectId()

        await self.collection.insert_one(data)
        return data

    async def update(self, _id, data):
        await self.collection.replace_one({
            "_id": maybe_object_id(_id)
        }, data)

        return data

    async def delete(self, _id):
        record = await self.find(_id=_id)
        if not record:
            raise RecordNotExistsError(self.collection.name, _id)
        await self.collection.remove(record["_id"], record)

    async def find(self, _id=None):
        return await self.collection.find_one({"_id": maybe_object_id(_id)})



import jwt
import logging

from flasky.errors import FlaskyTornError, InvalidTokenError

logger = logging.getLogger("utilities.security")


class AgentIsNotAuthorized(FlaskyTornError):

    def __init__(self, message="Agent is not authorized for this action"):
        super().__init__(status_code=403, message=message)


class SecurityManager(object):

    def __init__(self, app, secret=None, issuer=None):
        self._app = app
        self._secret = secret or app.settings.get('secret', None)
        self._issuer = issuer
        self._app.before_request_funcs.insert(0, self._before_request_hook)
        self._user_loader = None

    def user_loader(self, f):
        self._user_loader = f
        return f

    async def _before_request_hook(self, handler, method_definition):
        required_permission = method_definition.get('secure', None)
        if not required_permission:
            return

        auth_header = handler.request.headers.get('Authorization', None)
        if not auth_header:
            raise AgentIsNotAuthorized(
                message="Authorization header is not found")

        if not auth_header.startswith("Bearer:"):
            raise AgentIsNotAuthorized(
                message="Authorization token is not Bearer type")

        token = auth_header.split(':')[1]
        try:
            decoded = jwt.decode(token.strip(),
                                 key=self._secret,
                                 verify=True,
                                 algorithms="HS256",
                                 )
            handler.context.token = decoded

            if self._user_loader:
                logger.info("User loader has been found. Loading user...")
                handler.context.user = await self._user_loader(handler, method_definition, decoded)

            user_permissions = decoded.get("roles", None)
            for user_permission in user_permissions:
                if implies(user_permission, required_permission):
                    logger.info(
                        "User<{}> has permission<{}> to execute this action."
                        .format(decoded["prn"], required_permission))
                    return

            logger.info(
                    "User<username={}, permissions={}> doesn't has "
                    "required permission<{}> to execute this action"
                    .format(
                        decoded["prn"], user_permissions, required_permission))

            raise AgentIsNotAuthorized()



        except jwt.exceptions.InvalidTokenError as e:
            raise InvalidTokenError(e.args[0])


_PART_DIVIDER = ":"
_SUBPART_DIVIDER = ","
_WILDCARD_TOKEN = "*"


def partify(permission_string):
    """
    @type permission_string: str
    """
    if not permission_string:
        raise ValueError("Wildcard string cannot be none or empty")
    permission_string = permission_string.strip()

    _parts = []

    splitted_parts = permission_string.split(_PART_DIVIDER)
    for splitted_part in splitted_parts:
        subparts = splitted_part.lower().split(_SUBPART_DIVIDER)
        if not subparts:
            raise ValueError(
                    "Wildcard string cannot contains"
                    "parts with only dividers.")
        _parts.append(set(subparts))

    if not _parts:
        raise ValueError("Wildcard string cannot contain only dividers")

    return _parts


def implies(permission_1, permission_2):
    permission_parts = partify(permission_1)
    other_permission_parts = partify(permission_2)

    i = 0
    for other_permission_part in other_permission_parts:
        #: if this permission has less part than other permission,
        #: everything after the number of parts contained
        #: in this permission is implied.
        #: eg: com.admin implies com.admin.read
        if len(permission_parts) - 1 < i:
            return True
        elif _WILDCARD_TOKEN not in permission_parts[i] and \
                not permission_parts[i].issuperset(other_permission_part):
            return False
        i += 1

    for i in range(i, len(permission_parts)-1):
        if _WILDCARD_TOKEN not in permission_parts[i]:
            return False

    return True

def implies_any(permission_collection, permission):
    for _permission in permission_collection:
        if implies(_permission, permission):
            return True

    return False



# coding=utf-8
import json
from helpers import json_helpers, exceptions
from jsonschema import Draft4Validator, FormatChecker
import datetime
import logging


def validate_schema(schema, data):
    validator = Draft4Validator(
        schema,
        types={"datetime": datetime.datetime},
        format_checker=FormatChecker())

    errors = sorted(validator.iter_errors(data), key=lambda e: e.path)
    if errors:
        logging.warning(errors)
        raise exceptions.QError(
            400,
            message="errors.validationError",
            errors=format_validation_errors(errors))


def check_schema(schema, data):
    validator = Draft4Validator(
        schema,
        types={"datetime": datetime.datetime},
        format_checker=FormatChecker())

    errors = sorted(validator.iter_errors(data), key=lambda e: e.path)
    if errors:
        logging.warning(errors)
    return True if not errors else False


def format_validation_errors(errors):
    messages = [{'.'.join(str(s) for s in list(x.path)): x.message}
                for x in errors]

    return messages
